set echo on

/*
tables we'll use to demonstrate the massive differences you can 
see between NOT IN and NOT exists and how to "tame" not in to be 
a better "not exists"!
*/
drop table t1;
drop table t2;

create table t1 as select * from all_objects where rownum <= 10000;
create table t2 as select * from all_objects where rownum <= 9950;
create index t2_idx on t2(object_id);
alter table t2 modify object_id not null;
pause

/* 
Watch as I dance on the stage -- as this query takes a truly long time!
*/
alter session set sql_trace=true;
select /*+ RULE */ count(*) 
  from t1 rbo
 where object_id not in ( select object_id from t2 );

select /*+ RULE */ count(*) 
  from t1 rbo
 where NOT EXISTS (select null from t2 where t2.object_id = rbo.object_id )
/
select /*+ RULE */ count(*) 
  from t1, t2 rbo
 where t1.object_id = rbo.object_id(+)
   and rbo.object_id IS NULL
/

alter session set sql_trace=false;
pause

/*
And now, we'll give the CBO a crack at the SAME queries
*/

analyze table t1 compute statistics
for table
for all indexes
for all indexed columns;

analyze table t2 compute statistics
for table 
for all indexes
for all indexed columns;

alter session set sql_trace=true;

/*
Now, don't blink or you'll miss the next query
*/
pause

select count(*) 
  from t1 cbo
 where object_id not in ( select object_id from t2 where object_id is not null );

select count(*) 
  from t1 cbo
 where NOT EXISTS (select null from t2 where t2.object_id = cbo.object_id )
/

select count(*) 
  from t1, t2 cbo
 where t1.object_id = cbo.object_id(+)
   and cbo.object_id IS NULL
/
@connect /
host tk
